@@ -7,7 +7,7 @@
  *	AccessExclusiveLocks and starting snapshots for Hot Standby mode.
  *	Plus conflict recovery processing.
  *
- * Portions Copyright (c) 1996-2011, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2012, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
@@ -28,6 +28,7 @@
 #include "storage/sinvaladt.h"
 #include "storage/standby.h"
 #include "utils/ps_status.h"
+#include "utils/timestamp.h"
 
 /* User-settable GUC parameters */
 int			vacuum_defer_cleanup_age;
@@ -201,7 +202,7 @@ ResolveRecoveryConflictWithVirtualXIDs(VirtualTransactionId *waitlist,
 		standbyWait_us = STANDBY_INITIAL_WAIT_US;
 
 		/* wait until the virtual xid is gone */
-		while (!ConditionalVirtualXactLockTableWait(*waitlist))
+		while (!VirtualXactLock(*waitlist, false))
 		{
 			/*
 			 * Report via ps if we have been waiting for more than 500 msec
@@ -282,7 +283,7 @@ ResolveRecoveryConflictWithTablespace(Oid tsid)
 	VirtualTransactionId *temp_file_users;
 
 	/*
-	 * Standby users may be currently using this tablespace for for their
+	 * Standby users may be currently using this tablespace for their
 	 * temporary files. We only care about current users because
 	 * temp_tablespace parameter will just ignore tablespaces that no longer
 	 * exist.
@@ -466,7 +467,7 @@ SendRecoveryConflictWithBufferPin(ProcSignalReason reason)
  * determine whether an actual deadlock condition is present: the lock we
  * need to wait for might be unrelated to any held by the Startup process.
  * Sooner or later, this mechanism should get ripped out in favor of somehow
- * accounting for buffer locks in DeadLockCheck().  However, errors here
+ * accounting for buffer locks in DeadLockCheck().	However, errors here
  * seem to be very low-probability in practice, so for now it's not worth
  * the trouble.
  */
@@ -657,7 +658,7 @@ StandbyReleaseOldLocks(int nxids, TransactionId *xids)
 	for (cell = list_head(RecoveryLockList); cell; cell = next)
 	{
 		xl_standby_lock *lock = (xl_standby_lock *) lfirst(cell);
-		bool	remove = false;
+		bool		remove = false;
 
 		next = lnext(cell);
 
@@ -667,8 +668,8 @@ StandbyReleaseOldLocks(int nxids, TransactionId *xids)
 			remove = false;
 		else
 		{
-			int		i;
-			bool	found = false;
+			int			i;
+			bool		found = false;
 
 			for (i = 0; i < nxids; i++)
 			{
@@ -717,6 +718,9 @@ standby_redo(XLogRecPtr lsn, XLogRecord *record)
 {
 	uint8		info = record->xl_info & ~XLR_INFO_MASK;
 
+	/* Backup blocks are not used in standby records */
+	Assert(!(record->xl_info & XLR_BKP_BLOCK_MASK));
+
 	/* Do nothing if we're not in hot standby mode */
 	if (standbyState == STANDBY_DISABLED)
 		return;
@@ -746,7 +750,7 @@ standby_redo(XLogRecPtr lsn, XLogRecord *record)
 		ProcArrayApplyRecoveryInfo(&running);
 	}
 	else
-		elog(PANIC, "relation_redo: unknown op code %u", info);
+		elog(PANIC, "standby_redo: unknown op code %u", info);
 }
 
 static void
@@ -1005,8 +1009,8 @@ LogAccessExclusiveLockPrepare(void)
 	 * RecordTransactionAbort() do not optimise away the transaction
 	 * completion record which recovery relies upon to release locks. It's a
 	 * hack, but for a corner case not worth adding code for into the main
-	 * commit path. Second, must must assign an xid before the lock is
-	 * recorded in shared memory, otherwise a concurrently executing
+	 * commit path. Second, we must assign an xid before the lock is recorded
+	 * in shared memory, otherwise a concurrently executing
 	 * GetRunningTransactionLocks() might see a lock associated with an
 	 * InvalidTransactionId which we later assert cannot happen.
 	 */
