@@ -21,12 +21,12 @@
  * lookup key's hash value as a partition number --- this will work because
  * of the way calc_bucket() maps hash values to bucket numbers.
  *
- * Portions Copyright (c) 1996-2010, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2012, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/utils/hash/dynahash.c,v 1.80 2010/01/02 16:57:56 momjian Exp $
+ *	  src/backend/utils/hash/dynahash.c
  *
  *-------------------------------------------------------------------------
  */
@@ -63,8 +63,6 @@
 
 #include "postgres.h"
 
-#include <limits.h>
-
 #include "access/xact.h"
 #include "storage/shmem.h"
 #include "storage/spin.h"
@@ -162,6 +160,7 @@ struct HTAB
 	MemoryContext hcxt;			/* memory context if default allocator used */
 	char	   *tabname;		/* table name (for error messages) */
 	bool		isshared;		/* true if table is in shared memory */
+	bool		isfixed;		/* if true, don't enlarge */
 
 	/* freezing a shared table isn't allowed, so we can keep state here */
 	bool		frozen;			/* true = no more inserts allowed */
@@ -201,8 +200,6 @@ static void hdefault(HTAB *hashp);
 static int	choose_nelem_alloc(Size entrysize);
 static bool init_htab(HTAB *hashp, long nelem);
 static void hash_corrupted(HTAB *hashp);
-static long next_pow2_long(long num);
-static int	next_pow2_int(long num);
 static void register_seq_scan(HTAB *hashp);
 static void deregister_seq_scan(HTAB *hashp);
 static bool has_seq_scans(HTAB *hashp);
@@ -377,13 +374,8 @@ hash_create(const char *tabname, long nelem, HASHCTL *info, int flags)
 	{
 		/* Doesn't make sense to partition a local hash table */
 		Assert(flags & HASH_SHARED_MEM);
-
-		/*
-		 * The number of partitions had better be a power of 2. Also, it must
-		 * be less than INT_MAX (see init_htab()), so call the int version of
-		 * next_pow2.
-		 */
-		Assert(info->num_partitions == next_pow2_int(info->num_partitions));
+		/* # of partitions had better be a power of 2 */
+		Assert(info->num_partitions == (1L << my_log2(info->num_partitions)));
 
 		hctl->num_partitions = info->num_partitions;
 	}
@@ -444,6 +436,8 @@ hash_create(const char *tabname, long nelem, HASHCTL *info, int flags)
 					 errmsg("out of memory")));
 	}
 
+	if (flags & HASH_FIXED_SIZE)
+		hashp->isfixed = true;
 	return hashp;
 }
 
@@ -524,6 +518,7 @@ init_htab(HTAB *hashp, long nelem)
 {
 	HASHHDR    *hctl = hashp->hctl;
 	HASHSEGMENT *segp;
+	long		lnbuckets;
 	int			nbuckets;
 	int			nsegs;
 
@@ -538,7 +533,9 @@ init_htab(HTAB *hashp, long nelem)
 	 * number of buckets.  Allocate space for the next greater power of two
 	 * number of buckets
 	 */
-	nbuckets = next_pow2_int((nelem - 1) / hctl->ffactor + 1);
+	lnbuckets = (nelem - 1) / hctl->ffactor + 1;
+
+	nbuckets = 1 << my_log2(lnbuckets);
 
 	/*
 	 * In a partitioned table, nbuckets must be at least equal to
@@ -556,7 +553,7 @@ init_htab(HTAB *hashp, long nelem)
 	 * Figure number of directory segments needed, round up to a power of 2
 	 */
 	nsegs = (nbuckets - 1) / hctl->ssize + 1;
-	nsegs = next_pow2_int(nsegs);
+	nsegs = 1 << my_log2(nsegs);
 
 	/*
 	 * Make sure directory is big enough. If pre-allocated directory is too
@@ -626,9 +623,9 @@ hash_estimate_size(long num_entries, Size entrysize)
 				elementAllocCnt;
 
 	/* estimate number of buckets wanted */
-	nBuckets = next_pow2_long((num_entries - 1) / DEF_FFACTOR + 1);
+	nBuckets = 1L << my_log2((num_entries - 1) / DEF_FFACTOR + 1);
 	/* # of segments needed for nBuckets */
-	nSegments = next_pow2_long((nBuckets - 1) / DEF_SEGSIZE + 1);
+	nSegments = 1L << my_log2((nBuckets - 1) / DEF_SEGSIZE + 1);
 	/* directory entries */
 	nDirEntries = DEF_DIRSIZE;
 	while (nDirEntries < nSegments)
@@ -669,9 +666,9 @@ hash_select_dirsize(long num_entries)
 				nDirEntries;
 
 	/* estimate number of buckets wanted */
-	nBuckets = next_pow2_long((num_entries - 1) / DEF_FFACTOR + 1);
+	nBuckets = 1L << my_log2((num_entries - 1) / DEF_FFACTOR + 1);
 	/* # of segments needed for nBuckets */
-	nSegments = next_pow2_long((nBuckets - 1) / DEF_SEGSIZE + 1);
+	nSegments = 1L << my_log2((nBuckets - 1) / DEF_SEGSIZE + 1);
 	/* directory entries */
 	nDirEntries = DEF_DIRSIZE;
 	while (nDirEntries < nSegments)
@@ -823,27 +820,6 @@ hash_search_with_hash_value(HTAB *hashp,
 	hctl->accesses++;
 #endif
 
-	/*
-	 * If inserting, check if it is time to split a bucket.
-	 *
-	 * NOTE: failure to expand table is not a fatal error, it just means we
-	 * have to run at higher fill factor than we wanted.  However, if we're
-	 * using the palloc allocator then it will throw error anyway on
-	 * out-of-memory, so we must do this before modifying the table.
-	 */
-	if (action == HASH_ENTER || action == HASH_ENTER_NULL)
-	{
-		/*
-		 * Can't split if running in partitioned mode, nor if frozen, nor if
-		 * table is the subject of any active hash_seq_search scans.  Strange
-		 * order of these tests is to try to check cheaper conditions first.
-		 */
-		if (!IS_PARTITIONED(hctl) && !hashp->frozen &&
-			hctl->nentries / (long) (hctl->max_bucket + 1) >= hctl->ffactor &&
-			!has_seq_scans(hashp))
-			(void) expand_table(hashp);
-	}
-
 	/*
 	 * Do the initial lookup
 	 */
@@ -964,12 +940,24 @@ hash_search_with_hash_value(HTAB *hashp,
 			currBucket->hashvalue = hashvalue;
 			hashp->keycopy(ELEMENTKEY(currBucket), keyPtr, keysize);
 
+			/* caller is expected to fill the data field on return */
+
 			/*
-			 * Caller is expected to fill the data field on return.  DO NOT
-			 * insert any code that could possibly throw error here, as doing
-			 * so would leave the table entry incomplete and hence corrupt the
-			 * caller's data structure.
+			 * Check if it is time to split a bucket.  Can't split if running
+			 * in partitioned mode, nor if table is the subject of any active
+			 * hash_seq_search scans.  Strange order of these tests is to try
+			 * to check cheaper conditions first.
 			 */
+			if (!IS_PARTITIONED(hctl) &&
+			hctl->nentries / (long) (hctl->max_bucket + 1) >= hctl->ffactor &&
+				!has_seq_scans(hashp))
+			{
+				/*
+				 * NOTE: failure to expand table is not a fatal error, it just
+				 * means we have to run at higher fill factor than we wanted.
+				 */
+				expand_table(hashp);
+			}
 
 			return (void *) ELEMENTKEY(currBucket);
 	}
@@ -1349,6 +1337,9 @@ element_alloc(HTAB *hashp, int nelem)
 	HASHELEMENT *prevElement;
 	int			i;
 
+	if (hashp->isfixed)
+		return false;
+
 	/* Each element has a HASHELEMENT header plus user data. */
 	elementSize = MAXALIGN(sizeof(HASHELEMENT)) + MAXALIGN(hctlv->entrysize);
 
@@ -1403,32 +1394,11 @@ my_log2(long num)
 	int			i;
 	long		limit;
 
-	/* guard against too-large input, which would put us into infinite loop */
-	if (num > LONG_MAX / 2)
-		num = LONG_MAX / 2;
-
 	for (i = 0, limit = 1; limit < num; i++, limit <<= 1)
 		;
 	return i;
 }
 
-/* calculate first power of 2 >= num, bounded to what will fit in a long */
-static long
-next_pow2_long(long num)
-{
-	/* my_log2's internal range check is sufficient */
-	return 1L << my_log2(num);
-}
-
-/* calculate first power of 2 >= num, bounded to what will fit in an int */
-static int
-next_pow2_int(long num)
-{
-	if (num > INT_MAX / 2)
-		num = INT_MAX / 2;
-	return 1 << my_log2(num);
-}
-
 
 /************************* SEQ SCAN TRACKING ************************/
 
