@@ -1,7 +1,7 @@
 /*
  * psql - the PostgreSQL interactive terminal
  *
- * Copyright (c) 2000-2011, PostgreSQL Global Development Group
+ * Copyright (c) 2000-2012, PostgreSQL Global Development Group
  *
  * src/bin/psql/input.c
  */
@@ -11,7 +11,6 @@
 #include <unistd.h>
 #endif
 #include <fcntl.h>
-#include <limits.h>
 
 #include "input.h"
 #include "settings.h"
@@ -39,19 +38,14 @@ static int	history_lines_added;
  *	Preserve newlines in saved queries by mapping '\n' to NL_IN_HISTORY
  *
  *	It is assumed NL_IN_HISTORY will never be entered by the user
- *	nor appear inside a multi-byte string.  0x00 is not properly
+ *	nor appear inside a multi-byte string.	0x00 is not properly
  *	handled by the readline routines so it can not be used
  *	for this purpose.
  */
 #define NL_IN_HISTORY	0x01
 #endif
 
-#ifdef HAVE_ATEXIT
 static void finishInput(void);
-#else
-/* designed for use with on_exit() */
-static void finishInput(int, void *);
-#endif
 
 
 /*
@@ -163,7 +157,7 @@ pg_send_history(PQExpBuffer history_buf)
  *
  * Caller *must* have set up sigint_interrupt_jmp before calling.
  *
- * Note: we re-use a static PQExpBuffer for each call.  This is to avoid
+ * Note: we re-use a static PQExpBuffer for each call.	This is to avoid
  * leaking memory if interrupted by SIGINT.
  */
 char *
@@ -228,73 +222,23 @@ gets_fromFile(FILE *source)
 
 
 #ifdef USE_READLINE
-
-/*
- * Macros to iterate over each element of the history list in order
- *
- * You would think this would be simple enough, but in its inimitable fashion
- * libedit has managed to break it: in libreadline we must use next_history()
- * to go from oldest to newest, but in libedit we must use previous_history().
- * To detect what to do, we make a trial call of previous_history(): if it
- * fails, then either next_history() is what to use, or there's zero or one
- * history entry so that it doesn't matter which direction we go.
- *
- * In case that wasn't disgusting enough: the code below is not as obvious as
- * it might appear.  In some libedit releases history_set_pos(0) fails until
- * at least one add_history() call has been done.  This is not an issue for
- * printHistory() or encode_history(), which cannot be invoked before that has
- * happened.  In decode_history(), that's not so, and what actually happens is
- * that we are sitting on the newest entry to start with, previous_history()
- * fails, and we iterate over all the entries using next_history().  So the
- * decode_history() loop iterates over the entries in the wrong order when
- * using such a libedit release, and if there were another attempt to use
- * BEGIN_ITERATE_HISTORY() before some add_history() call had happened, it
- * wouldn't work.  Fortunately we don't care about either of those things.
- *
- * Usage pattern is:
- *
- *		BEGIN_ITERATE_HISTORY(varname);
- *		{
- *			loop body referencing varname->line;
- *		}
- *		END_ITERATE_HISTORY();
- */
-#define BEGIN_ITERATE_HISTORY(VARNAME) \
-	do { \
-		HIST_ENTRY *VARNAME; \
-		bool		use_prev_; \
-		\
-		history_set_pos(0); \
-		use_prev_ = (previous_history() != NULL); \
-		history_set_pos(0); \
-		for (VARNAME = current_history(); VARNAME != NULL; \
-			 VARNAME = use_prev_ ? previous_history() : next_history()) \
-		{ \
-			(void) 0
-
-#define END_ITERATE_HISTORY() \
-		} \
-	} while(0)
-
-
 /*
  * Convert newlines to NL_IN_HISTORY for safe saving in readline history file
  */
 static void
 encode_history(void)
 {
-	BEGIN_ITERATE_HISTORY(cur_hist);
-	{
-		char	   *cur_ptr;
+	HIST_ENTRY *cur_hist;
+	char	   *cur_ptr;
 
+	history_set_pos(0);
+	for (cur_hist = current_history(); cur_hist; cur_hist = next_history())
+	{
 		/* some platforms declare HIST_ENTRY.line as const char * */
 		for (cur_ptr = (char *) cur_hist->line; *cur_ptr; cur_ptr++)
-		{
 			if (*cur_ptr == '\n')
 				*cur_ptr = NL_IN_HISTORY;
-		}
 	}
-	END_ITERATE_HISTORY();
 }
 
 /*
@@ -303,18 +247,17 @@ encode_history(void)
 static void
 decode_history(void)
 {
-	BEGIN_ITERATE_HISTORY(cur_hist);
-	{
-		char	   *cur_ptr;
+	HIST_ENTRY *cur_hist;
+	char	   *cur_ptr;
 
+	history_set_pos(0);
+	for (cur_hist = current_history(); cur_hist; cur_hist = next_history())
+	{
 		/* some platforms declare HIST_ENTRY.line as const char * */
 		for (cur_ptr = (char *) cur_hist->line; *cur_ptr; cur_ptr++)
-		{
 			if (*cur_ptr == NL_IN_HISTORY)
 				*cur_ptr = '\n';
-		}
 	}
-	END_ITERATE_HISTORY();
 }
 #endif   /* USE_READLINE */
 
@@ -342,6 +285,16 @@ initializeInput(int flags)
 		history_lines_added = 0;
 
 		histfile = GetVariable(pset.vars, "HISTFILE");
+
+		if (histfile == NULL)
+		{
+			char	   *envhist;
+
+			envhist = getenv("PSQL_HISTORY");
+			if (envhist != NULL && strlen(envhist) > 0)
+				histfile = envhist;
+		}
+
 		if (histfile == NULL)
 		{
 			if (get_home_path(home))
@@ -365,24 +318,27 @@ initializeInput(int flags)
 	}
 #endif
 
-#ifdef HAVE_ATEXIT
 	atexit(finishInput);
-#else
-	on_exit(finishInput, NULL);
-#endif
 }
 
 
 /*
- * This function saves the readline history when psql exits.
+ * This function saves the readline history when user
+ * runs \s command or when psql exits.
  *
  * fname: pathname of history file.  (Should really be "const char *",
  * but some ancient versions of readline omit the const-decoration.)
  *
  * max_lines: if >= 0, limit history file to that many entries.
+ *
+ * appendFlag: if true, try to append just our new lines to the file.
+ * If false, write the whole available history.
+ *
+ * encodeFlag: whether to encode \n as \x01.  For \s calls we don't wish
+ * to do that, but must do so when saving the final history file.
  */
-static bool
-saveHistory(char *fname, int max_lines)
+bool
+saveHistory(char *fname, int max_lines, bool appendFlag, bool encodeFlag)
 {
 #ifdef USE_READLINE
 
@@ -392,15 +348,11 @@ saveHistory(char *fname, int max_lines)
 	 * where write_history will fail because it tries to chmod the target
 	 * file.
 	 */
-	if (strcmp(fname, DEVNULL) != 0)
+	if (useHistory && fname &&
+		strcmp(fname, DEVNULL) != 0)
 	{
-		/*
-		 * Encode \n, since otherwise readline will reload multiline history
-		 * entries as separate lines.  (libedit doesn't really need this, but
-		 * we do it anyway since it's too hard to tell which implementation we
-		 * are using.)
-		 */
-		encode_history();
+		if (encodeFlag)
+			encode_history();
 
 		/*
 		 * On newer versions of libreadline, truncate the history file as
@@ -414,6 +366,7 @@ saveHistory(char *fname, int max_lines)
 		 * see if the write failed.  Similarly for append_history.
 		 */
 #if defined(HAVE_HISTORY_TRUNCATE_FILE) && defined(HAVE_APPEND_HISTORY)
+		if (appendFlag)
 		{
 			int			nlines;
 			int			fd;
@@ -438,90 +391,33 @@ saveHistory(char *fname, int max_lines)
 			if (errno == 0)
 				return true;
 		}
-#else							/* don't have append support */
+		else
+#endif
 		{
 			/* truncate what we have ... */
 			if (max_lines >= 0)
 				stifle_history(max_lines);
-			/* ... and overwrite file.  Tough luck for concurrent sessions. */
+			/* ... and overwrite file.	Tough luck for concurrent sessions. */
 			errno = 0;
 			(void) write_history(fname);
 			if (errno == 0)
 				return true;
 		}
-#endif
 
 		psql_error("could not save history to file \"%s\": %s\n",
 				   fname, strerror(errno));
 	}
-#endif
-
-	return false;
-}
-
-
-/*
- * Print history to the specified file, or to the console if fname is NULL
- * (psql \s command)
- *
- * We used to use saveHistory() for this purpose, but that doesn't permit
- * use of a pager; moreover libedit's implementation behaves incompatibly
- * (preferring to encode its output) and may fail outright when the target
- * file is specified as /dev/tty.
- */
-bool
-printHistory(const char *fname, unsigned short int pager)
-{
-#ifdef USE_READLINE
-	FILE	   *output;
-	bool		is_pager;
-
-	if (!useHistory)
-		return false;
-
-	if (fname == NULL)
-	{
-		/* use pager, if enabled, when printing to console */
-		output = PageOutput(INT_MAX, pager);
-		is_pager = true;
-	}
-	else
-	{
-		output = fopen(fname, "w");
-		if (output == NULL)
-		{
-			psql_error("could not save history to file \"%s\": %s\n",
-					   fname, strerror(errno));
-			return false;
-		}
-		is_pager = false;
-	}
-
-	BEGIN_ITERATE_HISTORY(cur_hist);
-	{
-		fprintf(output, "%s\n", cur_hist->line);
-	}
-	END_ITERATE_HISTORY();
-
-	if (is_pager)
-		ClosePager(output);
-	else
-		fclose(output);
-
-	return true;
 #else
+	/* only get here in \s case, so complain */
 	psql_error("history is not supported by this installation\n");
-	return false;
 #endif
+
+	return false;
 }
 
 
 static void
-#ifdef HAVE_ATEXIT
 finishInput(void)
-#else
-finishInput(int exitstatus, void *arg)
-#endif
 {
 #ifdef USE_READLINE
 	if (useHistory && psql_history)
@@ -529,7 +425,7 @@ finishInput(int exitstatus, void *arg)
 		int			hist_size;
 
 		hist_size = GetVariableNum(pset.vars, "HISTSIZE", 500, -1, true);
-		(void) saveHistory(psql_history, hist_size);
+		saveHistory(psql_history, hist_size, true, true);
 		free(psql_history);
 		psql_history = NULL;
 	}
